//
//  WaveformChecker.m
//  Waveform Checker
//
//  Created by Steffen Kamp on Sun Jan 20 2002.
//  Copyright (c) 2001 __MyCompanyName__. All rights reserved.
//

#import "WaveformChecker.h"
#import <AudioUnit/AudioUnitProperties.h>



static const double	kAmplitude = 0.25;
static const double	kToneFrequency = 440.;

UInt32			gSinWaveFrameCount = 0; // this keeps track of the number of frames you render

// THESE values can be read from your data source
// they're used to tell the DefaultOutputUnit what you're giving it
#define GSAMPLERATE 44100
Float32			gSampleRate = (Float32)GSAMPLERATE;

UInt32 theFormatID = kAudioFormatLinearPCM;

UInt32 sampleNextPrinted = GSAMPLERATE;//(UInt32)gSampleRate;
UInt32 whenToPrint = (UInt32)(GSAMPLERATE / 1.0); // Once a second = 10== 10 times a second,

// data is 32 bit float
UInt32 theFormatFlags =  kLinearPCMFormatFlagIsFloat
#if BYTE_ORDER == BIG_ENDIAN
| kLinearPCMFormatFlagIsBigEndian
#endif
| kLinearPCMFormatFlagIsPacked;
UInt32 numBytesInAPacket = 8;
UInt32 numFramesPerPacket = 1; // this shouldn't change
UInt32 numBytesPerFrame = 8; //
UInt32 numChannelsPerFrame = 2;
UInt32 numBitsPerChannel = 32;



@implementation WaveformChecker

- (id)init {
    OSStatus err;

    struct AURenderCallbackStruct input;
    AudioStreamBasicDescription streamFormat;

    delegate = self;

    theOutputUnit = malloc(sizeof(AudioUnit));
    [self makeStandardSineWave];
    factor = 1.;

    // Open the default output unit
    verify_noerr(err = OpenADefaultComponent(kAudioUnitType_Output,
                                             kAudioUnitSubType_DefaultOutput,
                                             theOutputUnit));
    if (err) return nil;

    // Initialize it
    verify_noerr(AudioUnitInitialize(*theOutputUnit));

    // Set up a callback function to generate output to the output unit
    
    input.inputProc = MyRenderer;
    input.inputProcRefCon = self;

    verify_noerr(AudioUnitSetProperty(*theOutputUnit,
                                      kAudioUnitProperty_SetRenderCallback,
                                      kAudioUnitScope_Global,
                                      0,
                                      &input,
                                      sizeof(input)));

    // We tell the Output Unit what format we're going to supply data to it
    // this is necessary if you're providing data through an input callback
    // AND you want the DefaultOutputUnit to do any format conversions
    // necessary from your format to the device's format.
    
    streamFormat.mSampleRate = gSampleRate;		//	the sample rate of the audio stream
    streamFormat.mFormatID = theFormatID;			//	the specific encoding type of audio stream
    streamFormat.mFormatFlags = theFormatFlags;		//	flags specific to each format
    streamFormat.mBytesPerPacket = numBytesInAPacket;
    streamFormat.mFramesPerPacket = numFramesPerPacket;
    streamFormat.mBytesPerFrame = numBytesPerFrame;
    streamFormat.mChannelsPerFrame = numChannelsPerFrame;
    streamFormat.mBitsPerChannel = numBitsPerChannel;

    verify_noerr(AudioUnitSetProperty(
                                      *theOutputUnit,
                                      kAudioUnitProperty_StreamFormat,
                                      kAudioUnitScope_Input,
                                      0,
                                      &streamFormat,
                                      sizeof(AudioStreamBasicDescription)));

    //printf("Rendering source at %f\n", gSampleRate);

    

    // RUNS FOREVER!!!!
    // we call the CFRunLoopRunInMode to service any notifications that the audio
    // system has to deal with
    //while (1)
     //   CFRunLoopRunInMode(kCFRunLoopDefaultMode, 1, false);

    

    
    return self;
}

- (void)dealloc {
    [self stop:self];
    // Clean up
    CloseComponent(*theOutputUnit);
    free(theOutputUnit);
    //[delegate release];
    [super dealloc];
}

- (Float32)amplitudeForPhase:(Float32)phase {
    //NSLog(@"%f", (float)sin(phase));
    //return sin(phase);
    return sin(phase)*.4 + /*sin(phase*2)*.4 +*/ ((phase<M_PI)?.2:-.2);
}

- (void)makeStandardSineWave {
    UInt32 i;

    for(i = 0; i < WAVEFORMLENGTH; ++i) {
        Float32 nextFloat = [self amplitudeForPhase:(Float32)i / WAVEFORMLENGTH * (M_PI * 2.0)];

        waveform[i] = nextFloat;

    }
}

- (void)RenderSinAsIntWithStartingFrameCount:(UInt32)startingFrameCount frames:(UInt32)nFrames buffer:(void*)dataPtr bufferSize:(UInt32)sizeOfData {
    Float32 *fp = (Float32 *)dataPtr;
    Float32 pos, val;
    static Float32 lastPhase = 0;
    double fac;
    double lr;
    double amp;
    int i;


    //double j = startingFrameCount;
    //double cycleLength = gSampleRate / kToneFrequency;


    if(aboutToStart) {
        startFrame = startingFrameCount;
        aboutToStart = NO;
    }
    for (i = 0; i < nFrames; ++i) {
        int time = startingFrameCount+i-startFrame;
        if(time % 64 == 0) {
            double t = (double)time/gSampleRate;
            fac = [delegate factorForTime:t];
            lr = [delegate positionForTime:t];
            amp = [delegate amplitudeForTime:t];
        }

        pos = lastPhase + (Float32)(i+1)/gSampleRate*440.*2.*M_PI*fac;
        while(pos>2*M_PI)
            pos-=M_PI*2;
        val = [delegate amplitudeForPhase:pos]*amp;
        fp[0] = val * ( lr / 2. + .5);
        fp[1] = val * (-lr / 2. + .5);
        fp += 2;
        
        /*j += 1.0;
        if (j > cycleLength)
            j -= cycleLength;*/
    }
    lastPhase = pos;

#if DO_PRINT
    if (startingFrameCount > sampleNextPrinted) {
        printf ("Current Slice: Size of Buffer=%ld, nFrames=%ld, startingFrameCount=%ld\n",
                sizeOfData, nFrames, startingFrameCount);
        sampleNextPrinted += whenToPrint;
    }
#endif

#if STYV
    add += 1.;
    if(add > 440.)
        add = 0.;
#endif
}

OSStatus	MyRenderer(void *inRefCon, AudioUnitRenderActionFlags *ioActionFlags,
                    const AudioTimeStamp *inTimeStamp, UInt32 inBusNumber,
                    UInt32 inNumberFrames, AudioBufferList *ioData)
{
    UInt32 nFrames = inNumberFrames;


    [(WaveformChecker*)inRefCon RenderSinAsIntWithStartingFrameCount:gSinWaveFrameCount frames:nFrames buffer:ioData->mBuffers[0].mData bufferSize:ioData->mBuffers[0].mDataByteSize];
    gSinWaveFrameCount += nFrames;
    //gSinWaveFrameCount %= WAVEFORMLENGTH;
    return noErr;
}

- (IBAction)start:(id)sender {
    // Start the rendering
    // The DefaultOutputUnit will do any format conversions to the format of the default device
    aboutToStart = YES;
    verify_noerr(AudioOutputUnitStart(*theOutputUnit));
}

- (IBAction)stop:(id)sender {
    // REALLY after you're finished playing STOP THE AUDIO OUTPUT UNIT!!!!!!
    // but we never get here because we're running until the process is nuked...
    verify_noerr(AudioOutputUnitStop(*theOutputUnit));
}

- (IBAction)setFactor:(id)sender {
    factor = [sender floatValue];
}

/*- (double)getFactorForTime:(double)time {
    if(delegate!=nil)
        return [delegate factorForTime:time];

    return factor;
}

- (double)getPositionForTime:(double)time {
    if(delegate!=nil)
        return [delegate positionForTime:time];

    return 0.;
}

- (double)getAmplitudeForTime:(double)t {
    if(delegate!=nil)
        return [delegate amplitudeForTime:t];

    return 1.;
}*/

- (void)setDelegate:(id)del {
    if(del == nil)
        delegate = self;
    else
        delegate = del;
}

- (double)factorForTime:(double)time {
    return 1.;
}

- (double)positionForTime:(double)time {
    return 0.;
}

- (double)amplitudeForTime:(double)t {
    return 1.;
}

@end
